# -*- coding: utf-8 -*-
"""
Decides if vendor bundles are used or not.
Setup python path accordingly.
"""

from __future__ import absolute_import, print_function
import os.path
import sys

# -----------------------------------------------------------------------------
# DEFINES:
# -----------------------------------------------------------------------------
HERE = os.path.dirname(__file__)
TASKS_VENDOR_DIR = os.path.join(HERE, "_vendor")
INVOKE_BUNDLE = os.path.join(TASKS_VENDOR_DIR, "invoke.zip")
INVOKE_BUNDLE_VERSION = "0.13.0"

DEBUG_SYSPATH = False


# -----------------------------------------------------------------------------
# EXCEPTIONS:
# -----------------------------------------------------------------------------
class VersionRequirementError(SystemExit): pass

# -----------------------------------------------------------------------------
# FUNCTIONS:
# -----------------------------------------------------------------------------
def setup_path(invoke_minversion=None):
    """Setup python search and add ``TASKS_VENDOR_DIR`` (if available)."""
    # print("INVOKE.tasks: setup_path")
    if not os.path.isdir(TASKS_VENDOR_DIR):
        print("SKIP: TASKS_VENDOR_DIR=%s is missing" % TASKS_VENDOR_DIR)
        return
    elif os.path.abspath(TASKS_VENDOR_DIR) in sys.path:
        # -- SETUP ALREADY DONE:
        # return
        pass

    use_vendor_bundles = os.environ.get("INVOKE_TASKS_USE_VENDOR_BUNDLES", "no")
    if need_vendor_bundles(invoke_minversion):
        use_vendor_bundles = "yes"

    if use_vendor_bundles == "yes":
        syspath_insert(0, os.path.abspath(TASKS_VENDOR_DIR))
        if setup_path_for_bundle(INVOKE_BUNDLE, pos=1):
            import invoke
            bundle_path = os.path.relpath(INVOKE_BUNDLE, os.getcwd())
            print("USING: %s (version: %s)" % (bundle_path, invoke.__version__))
    else:
        # -- BEST-EFFORT: May rescue something
        syspath_append(os.path.abspath(TASKS_VENDOR_DIR))
        setup_path_for_bundle(INVOKE_BUNDLE, pos=len(sys.path))

    if DEBUG_SYSPATH:
        for index, p in enumerate(sys.path):
            print("  %d.  %s" % (index, p))


def require_invoke_minversion(min_version, verbose=False):
    """Ensures that :mod:`invoke` has at the least the :param:`min_version`.
    Otherwise,

    :param min_version: Minimal acceptable invoke version (as string).
    :param verbose:     Indicates if invoke.version should be shown.
    :raises: VersionRequirementError=SystemExit if requirement fails.
    """
    # -- REQUIRES: sys.path is setup and contains invoke
    try:
        import invoke
        invoke_version = invoke.__version__
    except ImportError:
        invoke_version = "__NOT_INSTALLED"

    if invoke_version < min_version:
        message = "REQUIRE: invoke.version >= %s (but was: %s)" % \
                  (min_version, invoke_version)
        message += "\nUSE: pip install invoke>=%s" % min_version
        raise VersionRequirementError(message)

    INVOKE_VERSION = os.environ.get("INVOKE_VERSION", None)
    if verbose and not INVOKE_VERSION:
        os.environ["INVOKE_VERSION"] = invoke_version
        print("USING: invoke.version=%s" % invoke_version)

def need_vendor_bundles(invoke_minversion=None):
    invoke_minversion = invoke_minversion or "0.0.0"
    need_vendor_answers = []
    need_vendor_answers.append(need_vendor_bundle_invoke(invoke_minversion))
    # -- REQUIRE: path.py
    try:
        import path
        need_bundle = False
    except ImportError:
        need_bundle = True
    need_vendor_answers.append(need_bundle)

    # -- DIAG: print("INVOKE: need_bundle=%s" % need_bundle1)
    # return need_bundle1 or need_bundle2
    return any(need_vendor_answers)

def need_vendor_bundle_invoke(invoke_minversion="0.0.0"):
    # -- REQUIRE: invoke
    try:
        import invoke
        need_bundle = invoke.__version__ < invoke_minversion
        if need_bundle:
            del sys.modules["invoke"]
            del invoke
    except ImportError:
        need_bundle = True
    except Exception:
        need_bundle = True
    return need_bundle

# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# -----------------------------------------------------------------------------
def setup_path_for_bundle(bundle_path, pos=0):
    if os.path.exists(bundle_path):
        syspath_insert(pos, os.path.abspath(bundle_path))
        return True
    return False

def syspath_insert(pos, path):
    if path in sys.path:
        sys.path.remove(path)
    sys.path.insert(pos, path)

def syspath_append(path):
    if path in sys.path:
        sys.path.remove(path)
    sys.path.append(path)




# -*- coding: UTF-8 -*-
"""
Test parameter related functionality.
"""

from __future__ import absolute_import, print_function
import os.path
from tests._test_support import write_configfile_with_contents
from click_configfile import Param, SectionSchema, ConfigFileReader, \
    assign_param_names, matches_section
import click
import pytest


# -----------------------------------------------------------------------------
# TEST CANDIDATE 1:
# -----------------------------------------------------------------------------
@assign_param_names
class ConfigSectionSchema1(object):

    @matches_section("hello")
    class Hello(SectionSchema):
        name    = Param(type=str)
        number  = Param(type=int, default=42)


class ConfigFileProcessor1(ConfigFileReader):
    config_files = ["hello.ini"]
    config_section_schemas = [
        ConfigSectionSchema1.Hello,
    ]


# -----------------------------------------------------------------------------
# TEST SUITE
# -----------------------------------------------------------------------------
xfail = pytest.mark.xfail

class TestCandidate1(object):

    def test_param_with_default__uses_param_default_when_missing(self, cli_runner_isolated):
        assert ConfigFileProcessor1.config_files[0] == "hello.ini"
        CONFIG_FILE_CONTENTS = """
            [hello]
            name = Alice
            """
        write_configfile_with_contents("hello.ini", CONFIG_FILE_CONTENTS)
        assert os.path.exists("hello.ini")

        CONTEXT_SETTINGS = dict(default_map=ConfigFileProcessor1.read_config())
        @click.command(context_settings=CONTEXT_SETTINGS)
        @click.option("-n", "--name", type=str, default="__CMDLINE__")
        @click.option("--number", default=123)
        def hello_with_param_default(name, number):
            click.echo("param: number= %s" % number)

        result = cli_runner_isolated.invoke(hello_with_param_default)
        assert result.output == "param: number= 42\n"
        assert result.exit_code == 0

    def test_param_with_default__uses_cmdline_when_provided(self, cli_runner_isolated):
        assert ConfigFileProcessor1.config_files[0] == "hello.ini"
        CONFIG_FILE_CONTENTS = """
            [hello]
            name = Alice
            number = 43
            """
        write_configfile_with_contents("hello.ini", CONFIG_FILE_CONTENTS)
        assert os.path.exists("hello.ini")

        CONTEXT_SETTINGS = dict(default_map=ConfigFileProcessor1.read_config())
        @click.command(context_settings=CONTEXT_SETTINGS)
        @click.option("-n", "--name", type=str, default="__CMDLINE__")
        @click.option("--number", default=123)
        def hello_with_param_default2(name, number):
            click.echo("param: number= %s" % number)

        result = cli_runner_isolated.invoke(hello_with_param_default2,
                                            ["--number", "234"])
        assert result.output == "param: number= 234\n"
        assert result.exit_code == 0

    def test_param_without_default__uses_cmdline_default_when_missing(self, cli_runner_isolated):
        assert ConfigFileProcessor1.config_files[0] == "hello.ini"
        CONFIG_FILE_CONTENTS = """
            [hello]
            number = 1234
            """
        write_configfile_with_contents("hello.ini", CONFIG_FILE_CONTENTS)
        assert os.path.exists("hello.ini")

        CONTEXT_SETTINGS = dict(default_map=ConfigFileProcessor1.read_config())
        @click.command(context_settings=CONTEXT_SETTINGS)
        @click.option("-n", "--name", type=str, default="__CMDLINE__")
        def hello_with_config(name):
            click.echo("param: name= %s" % name)

        result = cli_runner_isolated.invoke(hello_with_config)
        assert result.output == "param: name= __CMDLINE__\n"
        assert result.exit_code == 0




# -*- coding: UTF-8 -*-
"""
Provides tasks to build documentation with sphinx, etc.
"""

from __future__ import absolute_import, print_function
from invoke import task, Collection
from invoke.util import cd
from path import Path
import os.path
import sys

# -- TASK-LIBRARY:
from .clean import cleanup_tasks, cleanup_dirs

# -----------------------------------------------------------------------------
# TASKS:
# -----------------------------------------------------------------------------
@task
def clean(ctx, dry_run=False):
    """Cleanup generated document artifacts."""
    basedir = ctx.sphinx.destdir or "build/docs"
    cleanup_dirs([basedir], dry_run=dry_run)


@task(help={
    "builder": "Builder to use (html, ...)",
    "options": "Additional options for sphinx-build",
})
def build(ctx, builder="html", options=""):
    """Build docs with sphinx-build"""
    sourcedir = ctx.config.sphinx.sourcedir
    destdir = Path(ctx.config.sphinx.destdir or "build")/builder
    destdir = destdir.abspath()
    with cd(sourcedir):
        destdir_relative = Path(".").relpathto(destdir)
        command = "sphinx-build {opts} -b {builder} {sourcedir} {destdir}" \
                    .format(builder=builder, sourcedir=".",
                            destdir=destdir_relative, opts=options)
        ctx.run(command)

@task
def linkcheck(ctx):
    """Check if all links are corect."""
    build(ctx, builder="linkcheck")

@task
def browse(ctx):
    """Open documentation in web browser."""
    page_html = Path(ctx.config.sphinx.destdir)/"html"/"index.html"
    if not page_html.exists():
        build(ctx, builder="html")
    assert page_html.exists()
    open_cmd = "open"   # -- WORKS ON: MACOSX
    if sys.platform.startswith("win"):
        open_cmd = "start"
    ctx.run("{open} {page_html}".format(open=open_cmd, page_html=page_html))
    # ctx.run('python -m webbrowser -t {page_html}'.format(page_html=page_html))
    # -- DISABLED:
    # import webbrowser
    # print("Starting webbrowser with page=%s" % page_html)
    # webbrowser.open(str(page_html))

@task(help = {"dest": "Destination directory to save docs"})
def save(ctx, dest="docs.html", format="html"):
    """Save/update docs under destination directory."""
    print("STEP: Generate docs in HTML format")
    build(ctx, builder=format)

    print("STEP: Save docs under %s/" % dest)
    source_dir = Path(ctx.config.sphinx.destdir)/format
    Path(dest).rmtree_p()
    source_dir.copytree(dest)

    # -- POST-PROCESSING: Polish up.
    for part in [ ".buildinfo", ".doctrees" ]:
        partpath = Path(dest)/part
        if partpath.isdir():
            partpath.rmtree_p()
        elif partpath.exists():
            partpath.remove_p()

# -----------------------------------------------------------------------------
# TASK CONFIGURATION:
# -----------------------------------------------------------------------------
namespace = Collection(clean, linkcheck, browse, save)
namespace.add_task(build, default=True)
namespace.configure({
    "sphinx": {
        "sourcedir": "docs",
        "destdir": "build/docs"
    }
})

# -- ADD CLEANUP TASK:
cleanup_tasks.add_task(clean, "clean_docs")
cleanup_tasks.configure(namespace.configuration())



# -*- coding: UTF-8 -*-
"""
Pytest configuration file.
"""
from click.testing import CliRunner
import pytest

@pytest.fixture(scope="function")
def cli_runner(request):
    """click_ CLI runner to execute click_ commands.

    .. _click: https://click.pocoo.org/
    """
    return CliRunner()

@pytest.fixture(scope="function")
def cli_runner_isolated(request):
    """click_ CLI runner that provides an isolated filesystem.

    .. _click: https://click.pocoo.org/
    """
    cli_runner = CliRunner()
    with cli_runner.isolated_filesystem():
        yield cli_runner

@pytest.fixture(scope="function")
def isolated_filesystem(request):
    """click_ CLI runner that provides an isolated filesystem.

    .. _click: https://click.pocoo.org/
    """
    cli_runner = CliRunner()
    with cli_runner.isolated_filesystem() as _isolated_filesystem:
        yield _isolated_filesystem



# -*- coding: UTF-8 -*-
"""
Provides "invoke" script when invoke is not installed.
Note that this approach uses the "tasks/_vendor/invoke.zip" bundle package.

Usage::

    # -- INSTEAD OF: invoke command
    # Show invoke version
    python -m tasks --version

    # List all tasks
    python -m tasks -l

.. seealso::

    * http://pyinvoke.org
    * https://github.com/pyinvoke/invoke


Examples for Invoke Scripts using the Bundle
-------------------------------------------------------------------------------

For UNIX like platforms:

.. code-block:: sh

    #!/bin/sh
    #!/bin/bash
    # RUN INVOKE: From bundled ZIP file (with Bourne shell/bash script).
    # FILE: invoke.sh (in directory that contains tasks/ directory)

    HERE=$(dirname $0)
    export INVOKE_TASKS_USE_VENDOR_BUNDLES="yes"

    python ${HERE}/tasks/_vendor/invoke.zip $*


For Windows platform:

.. code-block:: bat

    @echo off
    REM RUN INVOKE: From bundled ZIP file (with Windows Batchfile).
    REM FILE: invoke.cmd (in directory that contains tasks/ directory)

    setlocal
    set HERE=%~dp0
    set INVOKE_TASKS_USE_VENDOR_BUNDLES="yes"
    if not defined PYTHON   set PYTHON=python

    %PYTHON% %HERE%tasks/_vendor/invoke.zip "%*"
"""

from __future__ import absolute_import
import os
import sys

# -----------------------------------------------------------------------------
# BOOTSTRAP PATH: Use provided vendor bundle if "invoke" is not installed
# -----------------------------------------------------------------------------
# NOTE: tasks/__init__.py performs sys.path setup.
os.environ["INVOKE_TASKS_USE_VENDOR_BUNDLES"] = "yes"

# -----------------------------------------------------------------------------
# AUTO-MAIN:
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    from invoke.main import program
    sys.exit(program.run())



# -*- coding: UTF-8 -*-
"""
Unit tests for :class:`click_configfile.ConfigFileReader`.
"""

from __future__ import absolute_import
import os.path


# -----------------------------------------------------------------------------
# TEST SUPPORT
# -----------------------------------------------------------------------------
def write_configfile_with_contents(filename, contents, encoding=None):
    encoding = encoding or "UTF-8"
    dirname = os.path.dirname(filename) or "."
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    # PREPARED: with open(filename, "w", encoding=encoding) as config_file:
    with open(filename, "w") as config_file:
        config_file.write(contents)



# -*- coding: UTF-8 -*-
"""
Invoke test tasks.
"""

from __future__ import print_function
from invoke import task, Collection
import os.path
import sys

# -- TASK-LIBRARY:
from .clean import cleanup_tasks, cleanup_dirs, cleanup_files


# ---------------------------------------------------------------------------
# TASKS
# ---------------------------------------------------------------------------
@task(name="all", help={
    "args": "Command line args for test run.",
})
def test_all(ctx, args="", options=""):
    """Run all tests (default)."""
    pytest_args = select_by_prefix(args, ctx.pytest.scopes)
    pytest_should_run = not args or (args and pytest_args)
    if pytest_should_run:
        pytest(ctx, pytest_args, options=options)

    behave_args = select_by_prefix(args, ctx.behave_test.scopes)
    behave_should_run = not args or (args and behave_args)
    if False and behave_should_run:
         behave(ctx, behave_args, options=options)


@task
def clean(ctx, dry_run=False):
    """Cleanup (temporary) test artifacts."""
    directories = ctx.test.clean.directories or []
    files = ctx.test.clean.files or []
    cleanup_dirs(directories, dry_run=dry_run)
    cleanup_files(files, dry_run=dry_run)


@task(name="unit")
def unittest(ctx, args="", options=""):
    """Run unit tests."""
    pytest(ctx, args, options)


@task
def pytest(ctx, args="", options=""):
    """Run unit tests."""
    args = args or ctx.pytest.args
    options = options or ctx.pytest.options
    ctx.run("pytest {options} {args}".format(options=options, args=args))


@task(help={
    "args": "Command line args for behave",
    "format": "Formatter to use (progress, pretty, ...)",
})
def behave(ctx, args="", format="", options=""):
    """Run behave tests."""
    format  = format or ctx.behave_test.format
    options = options or ctx.behave_test.options
    args = args or ctx.behave_test.args
    if os.path.exists("bin/behave"):
        behave = "{python} bin/behave".format(python=sys.executable)
    else:
        behave = "{python} -m behave".format(python=sys.executable)

    for group_args in grouped_by_prefix(args, ctx.behave_test.scopes):
        ctx.run("{behave} -f {format} {options} {args}".format(
            behave=behave, format=format, options=options, args=group_args))

@task(help={
    "args":     "Command line args for test",
    "report":   "Coverage report format to use (report, html, xml)",
})
def coverage(ctx, args="", report="report", append=False, combine=True):
    append = append or ctx.coverage.append
    combine = combine or ctx.coverage.combine
    report_formats = ctx.coverage.report_formats or []
    if report not in report_formats:
        report_formats.insert(0, report)
    opts = []
    if append:
        opts.append("--append")

    ctx.run("coverage run -m pytest {options} {args}".format(
            args=args, options=" ".join(opts)))
    if combine:
        ctx.run("coverage combine")
    for report_format in report_formats:
        ctx.run("coverage {report_format}".format(report_format=report_format))


# ---------------------------------------------------------------------------
# UTILITIES:
# ---------------------------------------------------------------------------
def select_prefix_for(arg, prefixes):
    for prefix in prefixes:
        if arg.startswith(prefix):
            return prefix
    return os.path.dirname(arg)

def select_by_prefix(args, prefixes):
    selected = []
    for arg in args.strip().split():
        assert not arg.startswith("-"), "REQUIRE: arg, not options"
        scope = select_prefix_for(arg, prefixes)
        if scope:
            selected.append(arg)
    return " ".join(selected)

def grouped_by_prefix(args, prefixes):
    """Group behave args by (directory) scope into multiple test-runs."""
    group_args = []
    current_scope = None
    for arg in args.strip().split():
        assert not arg.startswith("-"), "REQUIRE: arg, not options"
        scope = select_prefix_for(arg, prefixes)
        if scope != current_scope:
            if group_args:
                # -- DETECTED GROUP-END:
                yield " ".join(group_args)
                group_args = []
            current_scope = scope
        group_args.append(arg)
    if group_args:
        yield " ".join(group_args)


# ---------------------------------------------------------------------------
# TASK MANAGEMENT / CONFIGURATION
# ---------------------------------------------------------------------------
namespace = Collection(clean, unittest, pytest, coverage)
namespace.add_task(test_all, default=True)
namespace.configure({
    "test": {
        "clean": {
            "directories": [
                ".cache", "assets",                         # -- TEST RUNS
                "__WORKDIR__", "reports", "test_results",   # -- BEHAVE test
            ],
            "files": [
                ".coverage", ".coverage.*",
                "report.html",
                "rerun*.txt", "rerun*.featureset", "testrun*.json",
            ],
        },
    },
    "pytest": {
        "scopes":   ["tests"],
        "args":   "",
        "options": "",  # -- NOTE:  Overide in configfile "invoke.yaml"
    },
    # "behave_test": behave.namespace._configuration["behave_test"],
    "behave_test": {
        "scopes":   ["features", "issue.features"],
        "args":     "features issue.features",
        "format":   "progress",
        "options":  "",  # -- NOTE:  Overide in configfile "invoke.yaml"
    },
    "coverage": {
        "append":   False,
        "combine":  True,
        "report_formats": ["report", "html"],
    },
})

# -- ADD CLEANUP TASK:
cleanup_tasks.add_task(clean, "clean_test")
cleanup_tasks.configure(namespace.configuration())



# -*- coding: UTF-8 -*-
"""
Unit tests for :class:`click_configfile.ConfigFileReader`.
"""

from __future__ import absolute_import, print_function
import os.path
from tests._test_support import write_configfile_with_contents
from click_configfile import generate_configfile_names
import pytest


# -----------------------------------------------------------------------------
# TEST SUITE
# -----------------------------------------------------------------------------
class TestFunction(object):

    # -- TESTS FOR: generate_configfile_names()
    def test_generate_configfile_names__without_config_files(self, isolated_filesystem):
        config_files = list(generate_configfile_names([]))
        assert config_files == []

    def test_generate_configfile_names__provides_reversed_order(self, isolated_filesystem):
        # -- SETUP:
        EMPTY_CONTENTS = "# -- EMPTY\n"
        write_configfile_with_contents("hello.ini", EMPTY_CONTENTS)
        write_configfile_with_contents("hello.cfg", EMPTY_CONTENTS)
        assert os.path.exists("hello.ini")
        assert os.path.exists("hello.cfg")

        # -- PERFROM TEST:
        given_config_files = ["hello.ini", "hello.cfg"]
        actual_config_files = list(generate_configfile_names(given_config_files))
        expected_config_files = [
            os.path.join(".", "hello.cfg"),
            os.path.join(".", "hello.ini"),
        ]
        expected_config_files2 = [os.path.normpath(p)
                                  for p in expected_config_files]
        assert actual_config_files == expected_config_files
        assert list(reversed(given_config_files)) == expected_config_files2

    def test_generate_configfile_names__provides_existing_variants_with_searchpath(self, isolated_filesystem):
        # -- SETUP:
        EMPTY_CONTENTS = "# -- EMPTY\n"
        write_configfile_with_contents("hello.ini", EMPTY_CONTENTS)
        write_configfile_with_contents("more/hello.ini", EMPTY_CONTENTS)
        write_configfile_with_contents("more/hello.cfg", EMPTY_CONTENTS)
        assert os.path.exists("hello.ini")
        assert os.path.exists("more/hello.ini")
        assert os.path.exists("more/hello.cfg")

        # -- PERFROM TEST:
        given_config_files = ["hello.ini", "hello.cfg"]
        config_searchpath = [".", "more"]
        actual_config_files = list(generate_configfile_names(given_config_files,
                                                             config_searchpath))
        expected_config_files = [
            os.path.join("more", "hello.cfg"),
            os.path.join("more", "hello.ini"),
            os.path.join(".", "hello.ini"),
        ]
        assert actual_config_files == expected_config_files

    def test_generate_configfile_names__without_searchpath_uses_curdir(self, isolated_filesystem):
        # -- SETUP:
        EMPTY_CONTENTS = "# -- EMPTY\n"
        write_configfile_with_contents("hello.ini", EMPTY_CONTENTS)
        write_configfile_with_contents("hello.cfg", EMPTY_CONTENTS)
        assert os.path.exists("hello.ini")
        assert os.path.exists("hello.cfg")

        # -- PERFROM TEST:
        given_config_files = ["hello.ini", "hello.cfg"]
        actual_config_files = list(generate_configfile_names(given_config_files))
        expected_config_files = [
            os.path.join(".", "hello.cfg"),
            os.path.join(".", "hello.ini"),
        ]
        assert actual_config_files == expected_config_files

    def test_generate_configfile_names__missing_variants_are_removed(self, isolated_filesystem):
        # -- PRECONDITIONS:
        assert not os.path.exists("hello.ini")
        assert not os.path.exists("hello.cfg")

        # -- PERFROM TEST:
        given_config_files = ["hello.ini", "hello.cfg"]
        actual_config_files = list(generate_configfile_names(given_config_files))
        assert actual_config_files == []

    def test_generate_configfile_names__when_searchpath_part_isfile(self, isolated_filesystem):
        # -- SETUP:
        EMPTY_CONTENTS = "# -- EMPTY\n"
        write_configfile_with_contents("hello.ini", EMPTY_CONTENTS)
        write_configfile_with_contents("BAD_PART", EMPTY_CONTENTS)
        assert os.path.exists("hello.ini")
        assert os.path.exists("BAD_PART")

        # -- PERFROM TEST:
        given_config_files = ["hello.ini", "hello.cfg"]
        config_searchpath = [".", "BAD_PART"]
        actual_config_files = list(generate_configfile_names(given_config_files,
                                                             config_searchpath))
        expected_config_files = [
            os.path.join(".", "hello.ini"),
        ]
        assert actual_config_files == expected_config_files



# -*- coding: UTF-8 -*-
"""
Unit tests for config section schema description.
"""

from __future__ import absolute_import, print_function
from click_configfile import Param, SectionSchema
from click_configfile import assign_param_names, matches_section
import pytest


# -----------------------------------------------------------------------------
# TEST SUITE
# -----------------------------------------------------------------------------
class TestDecorators(object):

    # -- TESTS RELATED TO: @assign_param_names decorator
    def test_assign_param_names__with_schema_class(self):
        @assign_param_names
        class ExampleSchema(SectionSchema):
            some_non_param = 123
            person = Param(type=str)
            number = Param(type=int)
            verbose = Param(type=bool)
            persons = Param(type=str, multiple=True)

        # -- ENSURE:
        assert ExampleSchema.person.name == "person"
        assert ExampleSchema.number.name == "number"
        assert ExampleSchema.verbose.name == "verbose"
        assert ExampleSchema.persons.name == "persons"

    def test_assign_param_names__with_other_class(self):
        @assign_param_names
        class ExampleSchema(object):
            person = Param(type=str)
            number = Param(type=int)

        # -- ENSURE:
        assert ExampleSchema.person.name == "person"
        assert ExampleSchema.number.name == "number"

    def test_assign_param_names__with_nested_schema_class(self):
        @assign_param_names
        class ConfigSectionSchema(object):
            class Example(SectionSchema):
                person = Param(type=str)
                number = Param(type=int)

        # -- ENSURE:
        assert ConfigSectionSchema.Example.person.name == "person"
        assert ConfigSectionSchema.Example.number.name == "number"

    def test_assign_param_names__with_deeper_nested_schema_class(self):
        @assign_param_names
        class ConfigSectionSchema(object):
            class Level(object):
                class Example(SectionSchema):
                    person = Param(type=str)
                    number = Param(type=int)

        # -- ENSURE:
        assert ConfigSectionSchema.Level.Example.person.name == "person"
        assert ConfigSectionSchema.Level.Example.number.name == "number"

    def test_assign_param_names__with_derived_param_class_in_schema(self):
        class DerivedParam(Param):
            pass

        @assign_param_names
        class ExampleSchema(object):
            person = DerivedParam(type=str)
            number = DerivedParam(type=int)

        # -- ENSURE:
        assert ExampleSchema.person.name == "person"
        assert ExampleSchema.number.name == "number"

    def test_assign_param_names__with_own_param_class_as_marker(self):
        class MyParam(Param):
            pass

        @assign_param_names(param_class=MyParam)
        class ExampleSchema(object):
            person = Param(type=str)
            number = MyParam(type=int)

        # -- ENSURE:
        assert ExampleSchema.person.name is None
        assert ExampleSchema.number.name == "number"


    def test_assign_param_names__skips_non_param_objects(self):
        @assign_param_names
        class ExampleSchema(SectionSchema):
            some_non_param = 123
            param_class_alias = Param

        # -- ENSURE:
        assert not hasattr(ExampleSchema.some_non_param, "name")
        assert not hasattr(ExampleSchema.param_class_alias, "name")

    def test_assign_param_names__skips_params_with_assigned_names(self):
        @assign_param_names
        class ExampleSchema(SectionSchema):
            person = Param("xxx", type=str)
            number = Param(name="PHI", type=float)

        # -- ENSURE:
        assert ExampleSchema.person.name == "xxx"
        assert ExampleSchema.number.name == "PHI"


    # -- TESTS RELATED TO: @matches_section decorator
    @pytest.mark.parametrize("section_name", [
        "hello", "hello.*", "*.hello"
    ])
    def test_matches_section__with_arg(self, section_name):
        @matches_section(section_name)
        class Hello(SectionSchema):
            pass

        assert Hello.section_names == [section_name]

    @pytest.mark.parametrize("section_names", [
        ["hello", "hello.*"]
    ])
    def test_matches_section__with_multiple_args(self, section_names):
        @matches_section(section_names)
        class Hello(SectionSchema):
            pass

        assert isinstance(section_names, (list, tuple))
        assert Hello.section_names == section_names

    def test_matches_section__with_several_decorators(self):
        @matches_section("foo")
        @matches_section("bar.*")
        class Hello(SectionSchema):
            pass

        assert Hello.section_names == ["foo", "bar.*"]

    def test_matches_section__with_other_class(self):
        @matches_section("hello")
        class Hello(object):
            pass

        assert Hello.section_names == ["hello"]

    def test_matches_section__with_class_that_has_section_names(self):
        @matches_section("hello")
        class Hello(SectionSchema):
            section_names = ["foo"]

        assert Hello.section_names == ["hello", "foo"]

    @pytest.mark.parametrize("bad_section_name", [None, 123, False, 1.23])
    def test_matches_section__with_bad_arg(self, bad_section_name):
        with pytest.raises(ValueError) as e:
            @matches_section(bad_section_name)
            class Hello(SectionSchema):
                pass

        expected = "ValueError: %r (expected: string, strings)" % bad_section_name
        assert expected in str(e)
